@@ -58,6 +58,8 @@ static void glade_xml_finalize(GObject *object);
 static void glade_xml_build_interface(GladeXML *xml, GladeInterface *iface,
 				      const char *root);
 
+static GType glade_xml_real_lookup_type(GladeXML*self, const char *gtypename);
+
 /**
  * glade_xml_get_type:
  *
@@ -97,6 +99,8 @@ glade_xml_class_init (GladeXMLClass *class)
 
     G_OBJECT_CLASS(class)->finalize = glade_xml_finalize;
 
+    class->lookup_type = glade_xml_real_lookup_type;
+
     glade_xml_tree_id = g_quark_from_static_string(glade_xml_tree_key);
     glade_xml_name_id = g_quark_from_static_string(glade_xml_name_key);
     glade_xml_tooltips_id = g_quark_from_static_string(glade_xml_tooltips_key);
@@ -735,42 +739,41 @@ glade_xml_add_signals(GladeXML *xml, GtkWidget *w, GladeWidgetInfo *info)
 static void
 glade_xml_add_atk_actions(GladeXML *xml, GtkWidget *w, GladeWidgetInfo *info)
 {
-    gint i;
-    gint n;
-    AtkObject *atko = gtk_widget_get_accessible (w);
+    gint i, n_actions;
+    AtkObject *accessible;
     AtkAction *action;
 
-    if (info->n_atk_actions) {
-        if (!ATK_IS_ACTION (atko)) {
-            g_warning("widgets of type %s don't have actions, but one is specified",
+    if (info->n_atk_actions == 0)
+	return;
+
+    accessible = gtk_widget_get_accessible (w);
+
+    if (!ATK_IS_ACTION (accessible)) {
+	g_warning("widgets of type %s don't have actions, but one is specified",
 	  	  G_OBJECT_TYPE_NAME (w));
-	    return;
-        }
+	return;
+    }
 
-        action = ATK_ACTION (atko);
-        n = atk_action_get_n_actions (action);    
+    action = ATK_ACTION (accessible);
+    n_actions = atk_action_get_n_actions (action);    
     
-        for (i = 0; i < info->n_atk_actions; i++) {
-    	    GladeAtkActionInfo *action_info = &info->atk_actions[i];
-	    int j;
-	    for (j=0; j<n; ++j) {
-	        if (strcmp (atk_action_get_name (action, j),
-			    action_info->action_name) == 0) {
-	            break;
-	        }
-	    }
-	    if (j < n) {
-	        atk_action_set_description (ATK_ACTION (atko),
-					    j,
-					    action_info->description);	
-	    } else {
-		/* don't show a warning -- the action names might not
-                 * be registered if libgail hasn't been loaded */
+    for (i = 0; i < info->n_atk_actions; i++) {
+	GladeAtkActionInfo *action_info = &info->atk_actions[i];
+	int j;
 
-		/* g_warning("could not find action named %s", action_info->action_name); */
+	for (j = 0; j < n_actions; j++) {
+	    if (strcmp (atk_action_get_name (action, j),
+			action_info->action_name) == 0) {
+		break;
 	    }
-        }
-    }	
+	}
+	if (j < n_actions) {
+	    atk_action_set_description (action, j, action_info->description);
+	} else {
+	    /* don't show a warning -- the action names might not
+	     * be registered if libgail hasn't been loaded */
+	}
+    }
 }
 
 /* helper function for adding relations */
@@ -785,14 +788,19 @@ add_relation(AtkRelationSet *relations, AtkRelationType relation_type,
     if (relation) {
 	/* add new target accessible to relation */
 	GPtrArray* target_array = atk_relation_get_target (relation);
+	guint i;
 
-	g_ptr_array_remove (target_array, target_accessible);
+	/* first check if target occurs in array ... */
+	for (i = 0; i < target_array->len; i++)
+	    if (g_ptr_array_index(target_array, i) == target_accessible)
+		return;
+	g_object_ref(target_accessible);
 	g_ptr_array_add (target_array, target_accessible);
     } else {
 	/* the relation hasn't been created yet ... */
-	atk_relation_set_add (relations,
-			      atk_relation_new (&target_accessible, 1,
-						relation_type));
+	relation = atk_relation_new(&target_accessible, 1, relation_type);
+	atk_relation_set_add (relations, relation);
+	g_object_unref(relation);
     }
 }
 
@@ -801,8 +809,14 @@ static void
 glade_xml_add_atk_relations(GladeXML *xml, GtkWidget *w, GladeWidgetInfo *info)
 {
     gint i;
-    AtkObject *atko = gtk_widget_get_accessible (w);
-    AtkRelationSet *relations = atk_object_ref_relation_set (atko);
+    AtkObject *accessible;
+    AtkRelationSet *relations;
+
+    if (info->n_relations == 0)
+	return;
+
+    accessible = gtk_widget_get_accessible (w);
+    relations = atk_object_ref_relation_set (accessible);
     
     for (i = 0; i < info->n_relations; i++) {
 	GladeAtkRelationInfo *rinfo = &info->relations[i];
@@ -852,20 +866,23 @@ static void
 glade_xml_add_accessibility_info(GladeXML *xml, GtkWidget *w, GladeWidgetInfo *info)
 {
     gint i;
-    AtkObject *atko = gtk_widget_get_accessible (w);
+    AtkObject *accessible = gtk_widget_get_accessible (w);
     
     for (i = 0; i < info->n_atk_props; i++) {
 	GParamSpec *pspec;
 	GValue value = { 0 };
 
-	pspec = g_object_class_find_property(g_type_class_ref(ATK_TYPE_OBJECT),
+	pspec = g_object_class_find_property(G_OBJECT_GET_CLASS(accessible),
 					     info->atk_props[i].name);
 	if (!pspec) {
 	    g_warning("unknown property `%s' for class `%s'",
-		      info->atk_props[i].name, g_type_name(ATK_TYPE_OBJECT));
+		      info->atk_props[i].name, G_OBJECT_TYPE_NAME(accessible));
 	    continue;
-	} else if (glade_xml_set_value_from_string (xml, pspec, info->atk_props[i].value, &value)) {
-	    g_object_set_property (G_OBJECT (atko), info->atk_props[i].name, &value);
+	} else if (glade_xml_set_value_from_string (xml, pspec,
+						    info->atk_props[i].value,
+						    &value)) {
+	    g_object_set_property (G_OBJECT (accessible),
+				   info->atk_props[i].name, &value);
 	    g_value_unset (&value);
         }
 	
@@ -1195,6 +1212,7 @@ glade_register_widget(GType type,
 		      GladeFindInternalChildFunc find_internal_child)
 {
     GladeWidgetBuildData *data;
+    gpointer old_data;
 
     g_return_if_fail(g_type_is_a(type, GTK_TYPE_WIDGET));
 
@@ -1209,6 +1227,10 @@ glade_register_widget(GType type,
     data->build_children = build_children;
     data->find_internal_child = find_internal_child;
 
+    old_data = g_type_get_qdata (type, glade_build_data_id);
+    if (old_data)
+	    g_free (old_data);
+
     g_type_set_qdata(type, glade_build_data_id, data);
 }
 
@@ -1234,7 +1256,18 @@ get_build_data(GType type)
     build_data = g_type_get_qdata(type, glade_build_data_id);
     
     if (build_data)
-	return build_data;
+	return build_data;  /* specific BuildData found */
+
+    /* gtkmm derives it's own types, but they aren't registed with glade_register_widget(),
+     * so look at the base type.
+     * If 3rd party custom widgets haven't used glade_register_widget() then this might mean that
+     * libglade mistakenly uses the base type's functions. */
+     build_data = g_type_get_qdata(g_type_parent(type), glade_build_data_id);
+
+    if (build_data)
+	return build_data;  /* specific BuildData found */
+
+    /* No specifc BuildData found - use something generic: */
     if (g_type_is_a(type, GTK_TYPE_CONTAINER))
 	return &container_build_data;
     else
@@ -1796,19 +1829,21 @@ glade_standard_build_children(GladeXML *self, GtkWidget *parent,
 GtkWidget *
 glade_xml_build_widget(GladeXML *self, GladeWidgetInfo *info)
 {
-    GType type;
+    GType type = G_TYPE_INVALID;
     GtkWidget *ret;
     
     GLADE_NOTE(BUILD, g_message("Widget class: %s\tname: %s",
-				info->class, info->name));
-    if (!strcmp (info->class, "Custom")) {
+				info->classname, info->name));
+    if (!strcmp (info->classname, "Custom")) {
 	ret = custom_new (self, info);
     } else {
-	type = g_type_from_name(info->class);
+		/* Call GladeXml's lookup_type() virtual function to get the gtype: */
+		type = (* GLADE_XML_GET_CLASS(self)->lookup_type) (self, info->classname);
+
 	if (type == G_TYPE_INVALID) {
 	    char buf[50];
-	    g_warning("unknown widget class '%s'", info->class);
-	    g_snprintf(buf, 49, "[a %s]", info->class);
+	    g_warning("unknown widget class '%s'", info->classname);
+	    g_snprintf(buf, 49, "[a %s]", info->classname);
 	    ret = gtk_label_new(buf);
 	} else {
 	    ret = get_build_data(type)->new(self, type, info);
@@ -2039,3 +2074,14 @@ glade_xml_set_common_params(GladeXML *self, GtkWidget *widget,
     if (g_object_get_qdata(G_OBJECT(widget), visible_id))
 	gtk_widget_show(widget);
 }
+
+/**
+ * glade_xml_real_lookup_type
+ *
+ * Default implementation of the lookup_type() virtual function.
+ */
+static GType
+glade_xml_real_lookup_type(GladeXML*self, const char *gtypename)
+{
+  return g_type_from_name(gtypename);
+}
